#   -------------------------------------------------------------
#   Copyright (c) Microsoft Corporation. All rights reserved.
#   -------------------------------------------------------------
"""Utilities for Churn Prediction on Spark"""
from datetime import datetime, timedelta
from typing import List, Tuple, Union

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.window import Window


def add_churn_labels(
    features_df: DataFrame,
    future_login_df: DataFrame,
    id_cols: Union[List[str], Tuple[str]] = None,
    label_col: str = "Churn",
) -> DataFrame:
    """
    Add churn labels (churn == 1, not_churn = 0) to features_df.

    :param features_df:
    :param future_login_df: login data that does not overlap with the ones we used to extract the features
    :param id_cols:
    :param label_col:
    :return:
    """
    id_cols = id_cols if id_cols else ["EntityId", "TitleId"]

    # Active users: 0
    future_login_df = future_login_df.select(id_cols).distinct().withColumn(label_col, F.lit(0))

    # Churned users: 1
    if label_col in features_df.columns:
        features_df = features_df.drop(label_col)
    return features_df.join(future_login_df, on=id_cols, how="left").fillna(1)


def split_by_date(input_df: DataFrame, date_col: str, date: datetime.date) -> Tuple[DataFrame, DataFrame]:
    """Split a DataFrame into two by date"""
    return input_df.where(F.col(date_col) <= date), input_df.where(F.col(date_col) > date)


def _tumbling_window(
    input_df: DataFrame,
    snapshot_date: datetime.date,
    date_col: str,
    window_size: int,
    window_num: int,
    id_cols: List[str] = None,
    out_cols: List[str] = None,
    agg_cols: List[str] = None,
    agg_col_objs: Union[Tuple[Column], List[Column]] = None,
) -> DataFrame:
    """
    Extract tumbling window feature.

    :param input_df:
    :param snapshot_date:
    :param date_col:
    :param window_size: Window size in days.
    :param window_num: The number of windows.
    :param id_cols:
    :param out_cols: Output column names.
    :param agg_cols: Aggregation column names. E.g. ["sum(OrderTotal)", "count(OrderTotal)"]
    :param agg_col_objs: Aggregation columns. E.g. [F.sum("OrderTotal"), F.count("OrderTotal")]
    :return:
    """
    id_cols = id_cols or ["EntityId", "TitleId"]
    agg_cols = agg_cols or ["countDistinct(Timestamp)"]
    agg_col_objs = agg_col_objs or F.countDistinct("Timestamp")

    if out_cols is not None and len(out_cols) != len(agg_col_objs) * window_num:
        raise ValueError(
            "The number of output column names should be the same as the number of aggregation columns * windows"
        )

    # Compute window_start_date's offset 'days' since 1970 to compute weekly aggregate
    window_first_date = snapshot_date - timedelta(days=window_size * window_num - 1)
    start_day_offset = (window_first_date - datetime(1970, 1, 1).date()).days % window_size

    # Window start dates that will be translated to columns after pivot
    window_start_dates = [snapshot_date - timedelta(days=window_size * i - 1) for i in range(1, window_num + 1)]

    step = f"{window_size} days"
    window = F.window(F.col(date_col), windowDuration=step, slideDuration=step, startTime=f"{start_day_offset} days",)

    # Aggregate by windows
    df_windowed = input_df.groupBy(id_cols + [window]).agg(*agg_col_objs)

    df_windowed.cache()

    # Translate the start date of each window to columns
    return (
        df_windowed.withColumn("start", F.to_date(df_windowed.window.start))
        .groupby(id_cols)
        .pivot("start", values=window_start_dates)
        .agg(*[F.sum(col) for col in agg_cols])
        .fillna(0)
        .toDF(*(id_cols + out_cols))
    )


def extract_engagement(
    df_login: DataFrame,
    snapshot_date: datetime.date,
    date_col: str,
    window_size: int,
    window_num: int,
    id_cols: Union[Tuple[str], List[str]] = None,
    platforms: Union[str, Tuple[str], List[str]] = "all",
    use_running_sum: bool = True,
) -> DataFrame:
    """Extract engagement features including:
    - LoginCount{1-window_num}Week: Login count for each of past 1st-window_numth week.
    - LoginDays{1-window_num}Week: Login days for each of past 1st-window_numth week.
    - Sun, Mon, ...: Login day of week ratio.
    - AndroidDevice, ...: Login platform ratio.
    """
    id_cols = id_cols if id_cols else ["EntityId", "TitleId"]

    out_cols = [f"{col}{week}Week" for week in range(1, window_num + 1) for col in ["LoginCount", "LoginDays"]]

    df_login.cache()

    if platforms == "all":  # Use all in the data, this line is redundant
        platforms = sorted([row[0] for row in df_login.select("Platform").distinct().collect()])

    # Login weekly counts
    df_feat_engagement = _tumbling_window(
        input_df=df_login,
        snapshot_date=snapshot_date,
        date_col=date_col,
        window_size=window_size,
        window_num=window_num,
        id_cols=id_cols,
        out_cols=out_cols,
        agg_cols=["sum(Login_Count)", f"count({date_col})"],
        agg_col_objs=[
            F.sum("Login_Count"),
            F.countDistinct(date_col),  # "Login_Count" could have multiple records in a day
        ],
    )
    # Running sum of weeks
    df_feat_engagement.cache()
    if use_running_sum:
        for i in range(2, window_num + 1):
            df_feat_engagement = df_feat_engagement.withColumn(
                f"LoginCount{i}Week", F.col(f"LoginCount{i-1}Week") + F.col(f"LoginCount{i}Week")
            ).withColumn(f"LoginDays{i}Week", F.col(f"LoginDays{i-1}Week") + F.col(f"LoginDays{i}Week"))

    # Login day-of-week ratio
    dayofweek = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]
    df_day_of_week = (
        df_login.withColumn("DayOfWeek", F.date_format(date_col, "E"))
        .groupBy(id_cols + ["DayOfWeek"])
        .agg(F.sum("Login_Count").alias("count"))
        .withColumn("LoginDayOfWeekRatio", F.col("count") / F.sum("count").over(Window.partitionBy(id_cols)))
        .groupBy(id_cols)
        .pivot("DayOfWeek", values=dayofweek)
        .sum("LoginDayOfWeekRatio")
        .toDF(*(id_cols + [f"LoginDayOfWeekRatio{dow}" for dow in dayofweek]))  # Rename columns
    )

    df_feat_engagement = df_feat_engagement.join(df_day_of_week, on=id_cols, how="outer")

    # Login platform ratio
    if platforms is not None:
        df_platform = (
            df_login.groupBy(id_cols + ["Platform"])
            .agg(F.sum("Login_Count").alias("count"))
            .withColumn("PlatformRatio", F.col("count") / F.sum("count").over(Window.partitionBy(id_cols)))
            .groupBy(id_cols)
            .pivot("Platform", values=platforms)
            .sum("PlatformRatio")
            .toDF(*(id_cols + [f"LoginPlatformRatio{pl}" for pl in platforms]))  # Rename columns
        )

        df_feat_engagement = df_feat_engagement.join(df_platform, on=id_cols, how="outer")

    return df_feat_engagement.fillna(0)


def extract_real_money_purchase(
    df_login: DataFrame,
    df_purchase: DataFrame,
    snapshot_date: datetime.date,
    window_size: int,
    window_num: int,
    date_col: str,
    id_cols: Union[Tuple[str], List[str]] = None,
    use_running_sum: bool = True,
) -> DataFrame:
    """
    Aggregate Real Money Purchases from Raw Events

    :param df_login:
    :param df_purchase:
    :param snapshot_date:
    :param window_size:
    :param window_num:
    :param date_col:
    :param id_cols:
    :param use_running_sum:
    :return:
    """
    out_cols = [
        f"{col}{week}Week" for week in range(1, window_num + 1) for col in ["OrderSum", "OrderCount", "OrderDays"]
    ]
    id_cols = id_cols if id_cols else ["EntityId", "TitleId"]

    # Purchase weekly sum and counts
    df_feat_real_money_purchase = _tumbling_window(
        input_df=df_purchase,
        snapshot_date=snapshot_date,
        date_col=date_col,
        window_size=window_size,
        window_num=window_num,
        id_cols=id_cols,
        out_cols=out_cols,
        agg_cols=["sum(OrderTotal_Sum)", "sum(Order_Count)", f"count({date_col})"],
        agg_col_objs=[F.sum("OrderTotal_Sum"), F.sum("Order_Count"), F.countDistinct(date_col)],
    )
    # Running sum
    df_feat_real_money_purchase.cache()
    if use_running_sum:
        for i in range(2, window_num + 1):
            df_feat_real_money_purchase = (
                df_feat_real_money_purchase.withColumn(
                    f"OrderSum{i}Week", F.col(f"OrderSum{i-1}Week") + F.col(f"OrderSum{i}Week")
                )
                .withColumn(f"OrderCount{i}Week", F.col(f"OrderCount{i-1}Week") + F.col(f"OrderCount{i}Week"))
                .withColumn(f"OrderDays{i}Week", F.col(f"OrderDays{i-1}Week") + F.col(f"OrderDays{i}Week"))
            )

    return df_login.select(id_cols).distinct().join(df_feat_real_money_purchase, id_cols, "left").fillna(0)


def extract_tenure(
    df_title: DataFrame,
    df_login: DataFrame,
    snapshot_date: datetime.date,
    date_col: str = "Date",
    id_cols: List[str] = None,
) -> DataFrame:
    """Extract Tenure and DaysSinceLastLogin"""
    id_cols = id_cols if id_cols else ["EntityId", "TitleId"]

    df_title_add_date = (
        df_title.select(id_cols + [date_col]).groupBy(id_cols).agg(F.to_date(F.min(date_col)).alias("TitleAddDate"))
    )

    df_last_login = (
        df_login.select(id_cols + [date_col]).groupBy(id_cols).agg(F.to_date(F.max(date_col)).alias("LastLoginDate"))
    )

    return (
        df_title_add_date.join(df_last_login, id_cols, "inner")
        .withColumn("Tenure", F.datediff(F.lit(snapshot_date), "TitleAddDate"))
        .withColumn("DaysSinceLastLogin", F.datediff(F.lit(snapshot_date), "LastLoginDate"))
        .select(id_cols + ["Tenure", "DaysSinceLastLogin"])
    )


def get_labels(
    df_login: DataFrame,
    snapshot_date: datetime.date,
    date_col: str = "Date",
    id_cols: List[str] = None,
    label_col: str = "Churn",
    churn_window: int = 7,
) -> DataFrame:
    """
    New function to create labels from login data. The logic is to
    use the number of days since last login - counting from the
    snapshot_date and label as Churn if the days elapsed is greater than 6.
    """
    id_cols = id_cols if id_cols else ["EntityId", "TitleId"]

    df_last_login = (
        df_login.select(id_cols + [date_col]).groupBy(id_cols).agg(F.to_date(F.max(date_col)).alias("LastLoginDate"))
    )
    df_last_login = df_last_login.withColumn("DaysSinceLastLogin", F.datediff(F.lit(snapshot_date), "LastLoginDate"))
    df_last_login = df_last_login.withColumn(
        label_col, F.when(F.col("DaysSinceLastLogin") >= churn_window, 1).otherwise(0)
    )

    return df_last_login
