#   -------------------------------------------------------------
#   Copyright (c) Microsoft Corporation. All rights reserved.
#   -------------------------------------------------------------
"""Featurization Class for Churn Prediction on Spark"""
from abc import abstractmethod
from dataclasses import dataclass, field
from datetime import timedelta, date
from typing import List, Tuple

from pyspark.sql import SparkSession
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import to_date, expr, when, col, lit

from azure_gamedev.ai.playfab.utilities import _read_parquet_from_adls

KUSTO_DATASOURCE = "com.microsoft.kusto.spark.synapse.datasource"


@dataclass
class DeltaLakeStorageConfig:
    """Collect of Configurations related to Azure Storage Accounts"""

    storage_account_name: str
    input_data_container_name: str
    input_data_base_dir_name: str


@dataclass
class ChurnInputDataConfig:
    """Collect of Configurations related to File Paths"""

    base_input_path: str
    login_file_dir: str
    login_file_name: str
    title_file_dir: str
    title_file_name: str
    purchase_file_dir: str
    purchase_file_name: str


@dataclass
class ChurnWindowConfig:
    """Collection of Configurations related to Churn Window"""

    snapshot_date: date
    date_range: list
    churn_window: int


@dataclass
class ChurnLabelConfig:
    """Collection of Configurations related to Churn Labeling"""

    feature_and_label_window: int
    label_generation_method: int
    only_label_window1: int
    only_label_window2: int
    only_feature_window: int


@dataclass
class ChurnTitleConfig:
    """Collection of Configurations related to Churn Titles and Columns"""

    titles: str
    all_title_names: list
    col_title = "TitleId"
    col_date = "Timestamp"
    id_cols = ["EntityId", "TitleId"]
    missing_data: List[str] = field(default_factory=list)


@dataclass
class ChurnModelConfig:
    """Collection of Configs related to ML"""

    label_config: ChurnLabelConfig
    window_config: ChurnWindowConfig
    title_config: ChurnTitleConfig
    data_config: ChurnInputDataConfig


@dataclass
class KustoStorageConfig:
    """Kusto Storage Configuration"""

    linked_service: str
    kusto_database: str


class ChurnPrediction:
    """Basic Interface For Churn Prediction Methods"""

    @abstractmethod
    def get_num_days(self, task_type) -> int:
        """
        Use different window of days for different tasks.
        :param task_type:
        :return:
        """

    @abstractmethod
    def get_feature_date(self, task_type):
        """
        Determine feature Date by Task Type
        :param task_type: features_and_label, only_label, or only_features
        :return: date of features based on Task Type
        """

    @abstractmethod
    def get_player_added_title_df(self):
        """
        Load PlayFab Player Added Title Events to Spark DataFrame.

        Player Added Title Events indicate when a User first logged in.
        :return: Player Added Title Events
        """

    @abstractmethod
    def get_player_logged_in_df(self):
        """
        changed from zero-padded to non-zero-padded
        """

    @abstractmethod
    def get_player_realmoney_purchase_df(self):
        """
        Load PlayFab Player Real Money Purchase Events in Spark DataFrame.

        :return: PlayFab Player Real Money Purchase Events
        """

    @abstractmethod
    def merge_all_features(self, task_type, df_feat_tenure_recency, df_feat_engagement, df_feat_real_money_purchase):
        """
        Merge features for Tenure, Engagement and Purchases into a single DataFrame.

        :param task_type:
        :param df_feat_tenure_recency:
        :param df_feat_engagement:
        :param df_feat_real_money_purchase:
        :return:
        """


class DataLoader(ChurnPrediction):
    """Interface for Featurize with different input sources"""

    def __init__(self, config_json: dict):
        self.labels = ChurnLabelConfig(**config_json["label"])
        self.titles = ChurnTitleConfig(**config_json["title"])
        self.window = ChurnWindowConfig(**config_json["window"])
        self.data = ChurnInputDataConfig(**config_json["data"])

    def get_num_days(self, task_type: str) -> int:
        """
        Use different window of days for different tasks.
        :param task_type:
        :return:
        """
        if task_type == "features_and_label":
            return self.labels.feature_and_label_window
        if task_type == "only_features":
            return self.labels.only_feature_window
        if task_type == "only_label":
            return (
                self.labels.only_label_window1
                if self.labels.label_generation_method == 1
                else self.labels.only_label_window2
            )
        raise ValueError("Unknown Task Type")

    def get_feature_date(self, task_type: str) -> date:
        """
        Determine feature Date by Task Type
        :param task_type: features_and_label, only_label, or only_features
        :return: date of features based on Task Type
        """
        if (task_type == "features_and_label") or (
            task_type == "only_label" and self.labels.label_generation_method == 1
        ):
            return self.window.snapshot_date - timedelta(days=self.window.churn_window)
        if task_type == "only_features":
            return self.window.snapshot_date
        raise ValueError("Unknown Task Type")

    @abstractmethod
    def get_player_added_title_df(self):
        """
        Load PlayFab Player Added Title Events to Spark DataFrame.

        Player Added Title Events indicate when a User first logged in.
        :return: Player Added Title Events
        """

    @abstractmethod
    def get_player_logged_in_df(self):
        """
        changed from zero-padded to non-zero-padded
        """

    @abstractmethod
    def get_player_realmoney_purchase_df(self):
        """
        Load PlayFab Player Real Money Purchase Events in Spark DataFrame.

        :return: PlayFab Player Real Money Purchase Events
        """

    def merge_all_features(
        self, task_type: str, df_feat_tenure_recency, df_feat_engagement, df_feat_real_money_purchase
    ) -> DataFrame:
        """
        Merge features for Tenure, Engagement and Purchases into a single DataFrame.

        :param task_type:
        :param df_feat_tenure_recency:
        :param df_feat_engagement:
        :param df_feat_real_money_purchase:
        :return:
        """
        df_feat = None
        if (task_type in ["features_and_label", "only_features"]) or (
            task_type == "only_label" and self.labels.label_generation_method == 1
        ):
            df_feat = df_feat_tenure_recency.join(df_feat_engagement, on=self.titles.id_cols, how="inner")

        if task_type in ["features_and_label", "only_features"]:
            df_feat = df_feat.join(
                df_feat_real_money_purchase, on=self.titles.id_cols, how="inner"
            )  # inner is fine since all titles are present here
            # Add a flag to show whether realmoney_purchase data available or not
            df_feat = df_feat.withColumn(
                "PurchaseDataAvailable",
                when((col(self.titles.col_title).isin(self.titles.missing_data)), lit(0)).otherwise(lit(1)),
            )
        return df_feat


class DeltaLakeChurn(DataLoader):
    """Collection of Configurations and Data Loading Utilities for PlayFab Churn Featurization"""

    def __init__(self, config_json: dict = None):
        super().__init__(config_json)
        self.storage = DeltaLakeStorageConfig(**config_json["storage"])

    def get_player_added_title_df(self) -> DataFrame:
        """
        Load PlayFab Player Added Title Events to Spark DataFrame.

        Player Added Title Events indicate when a User first logged in.
        :return: Player Added Title Events
        """
        if self.titles.titles == "all":
            file_name = f"{self.data.title_file_name}/{self.data.title_file_dir}"
        else:
            file_name = [
                f"{self.data.title_file_name}/{self.data.title_file_dir}/{self.titles.col_title}={title}"
                for title in self.titles.all_title_names
            ]

        df_title = _read_parquet_from_adls(
            storage_name=self.storage.storage_account_name,
            container_name=self.storage.input_data_container_name,
            dir_name=self.storage.input_data_base_dir_name,
            file_name=file_name,
        )

        df_title = df_title.withColumn(self.titles.col_date, to_date("Timestamp"))
        return df_title

    def get_player_logged_in_df(self) -> DataFrame:
        """
        changed from zero-padded to non-zero-padded
        """
        file_path = f"{self.data.login_file_name}/{self.data.login_file_dir}"

        if self.titles.titles == "all":
            file_name = [f"{file_path}/Year={d.year}/Month={d.month}/Day={d.day}" for d in self.window.date_range]
        else:
            file_name = [
                f"{file_path}/Year={d.year}/Month={d.month}/Day={d.day}/{self.titles.col_title}={title}"
                for d in self.window.date_range
                for title in self.titles.all_title_names
            ]

        df_login = _read_parquet_from_adls(
            storage_name=self.storage.storage_account_name,
            container_name=self.storage.input_data_container_name,
            dir_name=self.storage.input_data_base_dir_name,
            file_name=file_name,
        )
        df_login = df_login.withColumn(self.titles.col_date, expr("make_date(year, month, day)"))
        return df_login

    def _check_realmoney_purchase_availability(self) -> Tuple[List[str], List[str]]:
        """Check whether real money purchase event is available for all titles"""
        available_data = []
        folder_names = [f"{self.titles.col_title}={title}" for title in self.titles.all_title_names]
        print("Checking for", folder_names)

        _path = f"{self.data.base_input_path}/{self.storage.input_data_base_dir_name}/"
        _purchase_dir = f"{self.data.purchase_file_name}/{self.data.purchase_file_dir}"
        file_names = [
            f"{_path}/{_purchase_dir}/Year={d.year}/Month={d.month}/Day={d.day}" for d in self.window.date_range
        ]
        data_dict = {k: [] for k in folder_names}
        for file in file_names:
            folders = mssparkutils.fs.ls(f"{file}")  # noqa: F821
            titles = [f.name for f in folders]
            for k in folder_names:
                if k in titles:
                    data_dict[k].append(file)

        missing_data = []
        for k in folder_names:
            if len(data_dict[k]) < len(file_names):
                missing_data.append(k.split("=")[-1])
            elif len(data_dict[k]) == len(file_names):
                available_data.append(k.split("=")[-1])

        print(f"Titles with real-money-purchase data: {available_data}")
        print(f"Missing data in {missing_data}")
        return available_data, missing_data

    def get_player_realmoney_purchase_df(self) -> DataFrame:
        """
        Load PlayFab Player Real Money Purchase Events in Spark DataFrame.

        :return: PlayFab Player Real Money Purchase Events
        """
        file_path = f"{self.data.purchase_file_name}/{self.data.purchase_file_dir}"

        if self.titles.titles == "all":
            file_name = [f"{file_path}/Year={d.year}/Month={d.month}/Day={d.day}" for d in self.window.date_range]
        else:
            available_data, _ = self._check_realmoney_purchase_availability()
            titles_with_data = [t for t in self.titles.all_title_names if t in available_data]
            file_name = [
                f"{file_path}/Year={d.year}/Month={d.month}/Day={d.day}/{self.titles.col_title}={title}"
                for d in self.window.date_range
                for title in titles_with_data
            ]

        df_purchase = _read_parquet_from_adls(
            storage_name=self.storage.storage_account_name,
            container_name=self.storage.input_data_container_name,
            dir_name=self.storage.input_data_base_dir_name,
            file_name=file_name,
        )
        df_purchase = df_purchase.withColumn(self.titles.col_date, expr("make_date(year, month, day)"))

        return df_purchase


class KustoChurn(DataLoader):
    """Churn Featurization Reading from Kusto"""

    def __init__(self, config_json: dict = None):
        super().__init__(config_json)
        self.storage = KustoStorageConfig(**config_json["storage"])

    def get_player_added_title_df(self) -> DataFrame:
        """
        Load PlayFab Player Added Title Events to Spark DataFrame.

        Player Added Title Events indicate when a User first logged in.
        :return: Player Added Title Events
        """
        query = f"""
        player_added_title
        | where
            ['time'] >= datetime_add('day', -{self.labels.only_feature_window}, datetime("{self.window.snapshot_date}"))
        and ['time'] < datetime_add('day', 1, datetime("{self.window.snapshot_date}"))
        and TitleId in (parse_csv("{','.join(self.titles.all_title_names)}"))
        | summarize hint.shufflekey=TitleId
            Timestamp=min(Timestamp)
        by EntityId, TitleId
        | project
            TitleId,
            EntityId,
            Timestamp,
            Year  = datetime_part("year", Timestamp),
            Month = datetime_part("month", Timestamp),
            Day   = datetime_part("day", Timestamp)
        """
        options = self._get_synapse_kusto(query)
        return (
            SparkSession.builder.getOrCreate()
            .read.format(options["format"])
            .load(**options)
            .withColumn(self.titles.col_date, expr("make_date(Year, Month, Day)"))
        )

    def get_player_logged_in_df(self) -> DataFrame:
        query = f"""player_logged_in
        | where
            ['time'] >= datetime_add('day', -{self.labels.only_feature_window}, datetime("{self.window.snapshot_date}"))
        and ['time'] < datetime_add('day', 1, datetime("{self.window.snapshot_date}"))
        and TitleId in (parse_csv("{','.join(self.titles.all_title_names)}"))
        | project
            TitleId, Platform, EntityId, Timestamp,
            Year  = datetime_part("year", Timestamp),
            Month = datetime_part("month", Timestamp),
            Day   = datetime_part("day", Timestamp)
        | summarize hint.strategy=shuffle
            Login_Count = count()
        by Year, Month, Day, TitleId, Platform, EntityId
        """
        options = self._get_synapse_kusto(query)
        return (
            SparkSession.builder.getOrCreate()
            .read.format(options["format"])
            .load(**options)
            .withColumn(self.titles.col_date, expr("make_date(Year, Month, Day)"))
            .withColumn("Timestamp", expr("make_date(Year, Month, Day)"))
            .select(
                "Timestamp",
                "Year",
                "Month",
                "Day",
                "TitleId",
                "EntityId",
                "Platform",
                "Login_Count",
                self.titles.col_date,
            )
        )

    def get_player_realmoney_purchase_df(self) -> DataFrame:
        query = f"""
        player_realmoney_purchase 
        | where
            ['time'] >= datetime_add('day', -{self.labels.only_feature_window}, datetime("{self.window.snapshot_date}"))
        and ['time'] < datetime_add('day', 1, datetime("{self.window.snapshot_date}"))
        and TitleId in (parse_csv("{','.join(self.titles.all_title_names)}"))
        | project
            TitleId, EntityId, OrderTotal, Timestamp,
            Year  = datetime_part("year", Timestamp),
            Month = datetime_part("month", Timestamp),
            Day   = datetime_part("day", Timestamp)
        | summarize hint.shufflekey=TitleId
            OrderTotal_Sum = sum(OrderTotal),
            Order_Count = count()
        by Year, Month, Day, TitleId, EntityId
        """
        options = self._get_synapse_kusto(query)
        return (
            SparkSession.builder.getOrCreate()
            .read.format(options["format"])
            .load(**options)
            .withColumn("Timestamp", expr("make_date(Year, Month, Day)"))
            .withColumn("Date", expr("make_date(Year, Month, Day)"))
        )

    def _get_synapse_kusto(self, query):
        return {
            "format": KUSTO_DATASOURCE,
            "spark.synapse.linkedService": self.storage.linked_service,
            "kustoDatabase": self.storage.kusto_database,
            "kustoQuery": query,
        }


class ChurnFeaturization(ChurnPrediction):
    """Churn Featurization"""

    def __init__(self, data_loader: DataLoader):
        self.featurization = data_loader

    def get_player_added_title_df(self) -> DataFrame:
        """
        Load PlayFab Player Added Title Events to Spark DataFrame.

        Player Added Title Events indicate when a User first logged in.
        :return: Player Added Title Events
        """
        return self.featurization.get_player_added_title_df()

    def get_player_logged_in_df(self) -> DataFrame:
        """
        changed from zero-padded to non-zero-padded
        """
        return self.featurization.get_player_logged_in_df()

    def get_player_realmoney_purchase_df(self) -> DataFrame:
        """
        Load PlayFab Player Real Money Purchase Events in Spark DataFrame.

        :return: PlayFab Player Real Money Purchase Events
        """
        return self.featurization.get_player_realmoney_purchase_df()

    @classmethod
    def create(cls, config_json: dict):
        """Create new instance of Churn Featurization"""
        if "storage_account_name" in config_json["storage"]:
            return ChurnFeaturization(DeltaLakeChurn(config_json))
        if "kusto_database" in config_json["storage"]:
            return ChurnFeaturization(KustoChurn(config_json))
        raise ValueError("Storage Type Not Supported")

    def merge_all_features(
        self, task_type: str, df_feat_tenure_recency, df_feat_engagement, df_feat_real_money_purchase
    ) -> DataFrame:
        """
        Merge features for Tenure, Engagement and Purchases into a single DataFrame.

        :param task_type:
        :param df_feat_tenure_recency:
        :param df_feat_engagement:
        :param df_feat_real_money_purchase:
        :return:
        """
        return self.featurization.merge_all_features(
            task_type, df_feat_tenure_recency, df_feat_engagement, df_feat_real_money_purchase
        )

    def get_num_days(self, task_type: str) -> int:
        return self.featurization.get_num_days(task_type)

    def get_feature_date(self, task_type: str) -> date:
        return self.featurization.get_feature_date(task_type)
